# -*- coding: utf-8 -*-

from .serializable import serialize, ATTR, LEN_ATTR
from .classes import Classes
from ..utils.common import lls
from ..dataset.metadata import guess_value
from ..dataset.finetime import FineTime

import logging
import json
import gzip
import binascii
import array
import urllib
from .odict import ODict
from collections.abc import MutableMapping as MM
import sys
if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = (str, bytes, bytearray)
else:
    PY3 = False
    strset = (str, unicode)

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

''' Note: this has to be in a different file where other interface
classes are defined to avoid circular dependency (such as ,
Serializable.
'''


def constructSerializable(obj, lookup=None, int_key=False, debug=False):
    """ mh: reconstruct object from the output of jason.loads().
    Recursively goes into nested class instances that are not
    encoded by default by JSONEncoder, instantiate and fill in
    variables.
    Objects to be deserialized must have their classes loaded.
    _STID cannot have module names in it (e.g.  dataset.Product)
    or locals()[classname] or globals()[classname] will not work.

    Parameters
    ----------
    :useint: is a key in key-value pair has only digits characters the key will be substituted by `int(key)`.
    Returns
    -------
    """
    global indent
    indent += 1
    spaces = '  ' * indent

    classname = obj.__class__.__name__
    if debug:
        print(spaces + '===OBJECT %s ===' % lls(obj, 150))
    if not hasattr(obj, '__iter__') or issubclass(obj.__class__, strset):
        if debug:
            print(spaces + 'Find non-iter <%s>' % classname)
        indent -= 1
        return obj

    # process list first
    if isinstance(obj, list):
        if debug:
            print(spaces + 'Find list <%s>' % classname)
        inst = []
        # loop i to preserve order
        for i in range(len(obj)):
            x = obj[i]
            xc = x.__class__
            if debug:
                print(spaces + 'looping through list %d <%s>' %
                      (i, xc.__name__))
            if issubclass(xc, (list, dict)):
                des = constructSerializable(
                    x, lookup=lookup, int_key=int_key, debug=debug)
            else:
                des = x
            inst.append(des)
        if debug:
            print(spaces + 'Done with list <%s>' % (classname))
        indent -= 1
        return inst

    if not '_STID' in obj:
        """ This object is supported by JSON encoder """
        if debug:
            print(spaces + 'Find non-_STID. <%s>' % classname)
        inst = obj
    else:
        ostid = obj['_STID']
        # check escape case
        if ostid.startswith('0'):
            # escape
            inst = obj
            # classname = ostid[1:]
            # inst['_STID'] = classname
            if debug:
                print(spaces + 'Find _STID <%s>. Take as <%s>.' %
                      (ostid, classname))
            indent -= 1
            return inst
        else:
            classname = ostid
            if debug:
                print(spaces + 'Find _STID <%s>.' % (ostid))
        # process types wrapped in a dict
        if PY3:
            if classname == 'datetime,tai':
                inst = FineTime(obj['code']).toDatetime()
                if debug:
                    print(spaces + 'Instanciate datetime')
                indent -= 1
                return inst
            if classname == 'bytes':
                inst = bytes.fromhex(obj['code'])
                # inst = codecs.decode(obj['code'], 'hex')
                if debug:
                    print(spaces + 'Instanciate bytes')
                indent -= 1
                return inst
            elif classname == 'bytes,gz' or classname == 'bytes,gz,b64':
                inst = gzip.decompress(binascii.a2b_base64(obj['code']))
                if debug:
                    print(spaces + 'Instanciate hex_gz')
                indent -= 1
                return inst
            elif classname == 'bytearray':
                inst = bytearray.fromhex(obj['code'])
                if debug:
                    print(spaces + 'Instanciate bytearray')
                indent -= 1
                return inst
            elif classname == 'bytearray,gz' or classname == 'bytearray,gz,b64':
                inst = bytearray(gzip.decompress(
                    binascii.a2b_base64(obj['code'])))
                if debug:
                    print(spaces + 'Instanciate bytearray_gz')
                indent -= 1
                return inst
            elif classname.startswith('a.array'):
                # format is "...a.array_I,..."
                array_t, tcode = tuple(classname.split(',', 1)[0].split('_'))
                if array_t == 'a.array':
                    inst = array.array(tcode, binascii.a2b_hex(obj['code']))
                    if debug:
                        print(spaces + 'Instanciate array.array')
                    indent -= 1
                    return inst
                elif array_t == 'a.array,gz,b64':
                    # gzip-base64
                    inst = array.array(tcode, gzip.decompress(
                        binascii.a2b_base64(obj['code'])))
                    if debug:
                        print(spaces + 'Instanciate array.array gzip base64')
                    indent -= 1
                    return inst
        if classname in lookup:
            # Now we have a blank instance.
            inst = lookup[classname]()
            if debug:
                print(spaces + 'Instanciate custom obj <%s>' % classname)
        elif classname == 'ellipsis':
            if debug:
                print(spaces + 'Instanciate Ellipsis')
            indent -= 1
            return Ellipsis
        elif classname in lookup and 'obj' in obj:
            o = constructSerializable(
                obj['obj'], lookup=lookup, int_key=int_key, debug=debug)
            inst = lookup[classname](o)
            obj = inst
            try:
                typ = obj.type
            except (LookupError, AttributeError) as e:
                typ = None
            if typ in ['image/svg']:
                ser = obj[ATTR+'data']
                ser = urllib.parse.unquote(ser)

            if debug:
                print(spaces + 'Instanciate defined %s' % obj['obj'])
            indent -= 1
            return inst
        elif classname == 'dtype':
            if debug:
                print(spaces + 'Instanciate type %s' % obj['obj'])
            inst = lookup[obj['obj']]
            if inst is None:
                __import__('pdb').set_trace()

            indent -= 1
            return inst
        else:
            raise ValueError('Class %s is not known.' % classname)
    if debug:
        print(spaces + 'Go through properties of instance')
    # we might change key during iteration so save the original keys
    for k in list(obj.keys()):
        """ loop through all key-value pairs. """
        v = obj[k]
        if k == '_STID':
            continue
        # deserialize v
        # should be object_pairs_hook in the following if... line
        if issubclass(v.__class__, (dict, list)):
            if debug:
                print(spaces + '[%s]value(dict/list) <%s>: %s' %
                      (k, v.__class__.__qualname__,
                       lls(list(iter(v)), 70)))
            desv = constructSerializable(
                v, lookup=lookup, int_key=int_key, debug=debug)
        else:
            if debug:
                print(spaces + '[%s]value(simple) <%s>: %s' %
                      (str(k), v.__class__.__name__, lls(v, 70)))
            if 1:
                desv = v
            else:
                if isinstance(v, str) or isinstance(v, bytes):
                    try:
                        desv = int(v)
                    except ValueError:
                        desv = v

        # set k with desv
        icn = inst.__class__.__name__
        dcn = desv.__class__.__name__
        if issubclass(inst.__class__, (MM)):    # should be object_pairs_hook
            # set attributes. JSON doesn't support attributes to all objects so we have to use the `ATTR` work-around. Because `inst` could actually could have a key starting with `ATTR`, we set the attributes only when inst is not `dict` or when '_STID' is a member.
            if issubclass(k.__class__, str) and k.startswith(ATTR) and (not isinstance(inst, dict) or '_STID' in inst):
                k2 = k[LEN_ATTR:]
                setattr(inst, k2, desv)
                if debug:
                    print(spaces + 'Set attrbute to dict/usrd <%s>.%s = %s <%s>' %
                          (icn, str(k2), lls(desv, 70), dcn))
            else:
                # convert key to int if required
                if int_key and k.isdigit():
                    # obj == inst so delete old entry if to change key
                    inst.pop(k)
                    inst[int(k)] = desv
                else:
                    inst[k] = desv
                if debug:
                    print(spaces + 'Set member to dict/usrd <%s>[%s] = %s <%s>' %
                          (icn, str(k), lls(desv, 70), dcn))
        else:
            if k.startswith(ATTR):
                k2 = k[LEN_ATTR:]
                setattr(inst, k2, desv)
                if debug:
                    print(spaces + 'set attribute to non-dict <%s>.%s = %s <%s>' %
                          (icn, str(k2), lls(desv, 70), dcn))
            else:
                setattr(inst, k, desv)
                if debug:
                    print(spaces + 'set attribute to non-dict <%s>.%s = %s <%s>' %
                          (icn, str(k), lls(desv, 70), dcn))
    indent -= 1
    return inst


class IntDecoder(json.JSONDecoder):
    """ adapted from https://stackoverflow.com/questions/45068797/how-to-convert-string-int-json-into-real-int-with-json-loads
    modified to also convert keys in dictionaries.
    """

    def decode(self, s):
        """
        Parameters
        ----------

        Returns
        -------
        """
        # result = super(Decoder, self).decode(s) for Python 2.x
        result = super(IntDecoder, self).decode(s)
        return self._decode(result)

    def _decode(self, o):
        """
        Parameters
        ----------

        Returns
        -------
        """
        if isinstance(o, str) or isinstance(o, bytes):
            try:
                return int(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return dict({self._decode(k): self._decode(v) for k, v in o.items()})
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o


class IntDecoderOD(IntDecoder):
    def _decode(self, o):
        """ Uses ODict
        Parameters
        ----------

        Returns
        -------
        """
        if isinstance(o, str) or isinstance(o, bytes):
            try:
                return int(o)
            except ValueError:
                return o
        elif isinstance(o, dict):
            return ODict({self._decode(k): self._decode(v) for k, v in o.items()})
        elif isinstance(o, list):
            return [self._decode(v) for v in o]
        else:
            return o


Class_Look_Up = Classes.mapping


def deserialize(js, lookup=None, debug=False, usedict=True, int_key=False, verbose=False):
    """ Create (Load) live object from string, bytes that are generated by serialization.



    Parameters
    ----------
    js : str, bytes.
         JSON string or bytes, made with `json.load`.
    lookup : mapping
         A `globals` or `locals` -like mapping that gives class object if the class name is given.
    debug : bool
         A if set `True`, print out step-by-step report of how a JSON document is deserialized into an object. Default `False`.
    usedict: bool
        If is `True` `dict` insted of `ODict` will be used during deserialization. Default `True`.
    init_key: bool
        If set, dictionary keys that are integers in string form, e.g. '1', are casted to integers. Default `False`.
    verbose: bool
        Give even more info.
    Returns
    -------
    object
       Deserialized object.
    """

    if lookup is None:
        lookup = Class_Look_Up

    if not isinstance(js, strset) or len(js) == 0:
        return None
    # debug = False  # True if issubclass(obj.__class__, list) else False
    try:
        if usedict:
            obj = json.loads(js)  # , cls=IntDecoder)
        else:
            # , cls=IntDecoderOD)
            obj = json.loads(js, object_pairs_hook=ODict)
    except json.decoder.JSONDecodeError as e:
        msg = 'Bad JSON====>\n%s\n<====\n%s' % (
            lls(js, 500), str(e))
        logging.error(msg)
        obj = msg
    if debug:
        # print('load-str ' + str(o) + ' class ' + str(o.__class__))
        print('-------- json loads returns: --------\n' + str(obj))

    global indent
    indent = -1
    return constructSerializable(obj, lookup=lookup, int_key=int_key, debug=debug)


Serialize_Args_Sep = '__'
SAS_Avatar = '~'


def encode_str(a0):
    """ quote to remove general url offenders then use a mostly harmless str to substitute Serialize_Args_Sep.
    """
    return urllib.parse.quote(a0).replace(Serialize_Args_Sep, SAS_Avatar)


def serialize_args(*args, not_quoted=False, **kwds):
    """
    Serialize all positional and keywords arguements as they would appear in a function call.
    Arguements are assumed to have been placed in the same order of a valid function/method call. They are scanned from left to right from `args[i]` i = 0, 1,... to `kwds[j]` j = 0, 1, ...

* Scan args from i=0. if is of args[i] is of `bool`, `int`, `float` types, convert with `str`, if `str()`, convert with `encode_str()`, if `bytes` or `bytearray' types, with ```0x```+`hex()`, save to the convered-list, and move on to the next element.
* else if finding a segment not of any of the above types,
** put this and the rest of ```args``` as the ```value``` in ```{'apiargs':value}```,
** and append `kwds` key-val pairs after this pair,
** serialize the disctionary with `serialize()` and encode_str()
** append the result to the converted-list.
** break from the args scan loop.
* if args scan loop reaches its end, if `kwds` is not empty, serialize it with `serialize()` and encode_str(),
or scanning reaches the end of args.
* append the result to the converted-list.
* join the converted-list with `Serialize_Args_Sep`.
* return the result string

    """
    noseriargs = []
    i = 0
    # print('AR ', args, ' KW ', kwds)
    # from ..pal.query import AbstractQuery
    # if len(args) and issubclass(args[0].__class__, AbstractQuery):
    #    __import__('pdb').set_trace()

    for i, a0 in enumerate(args):
        # a string or number or boolean
        a0c = a0.__class__
        if a0 is None or issubclass(a0c, (bool, int, float)):
            noseriargs.append(str(a0))
        elif issubclass(a0c, (str)):
            noseriargs.append(a0 if not_quoted else encode_str(a0))
        elif issubclass(a0c, (bytes, bytearray)):
            noseriargs.append('0x'+a0.hex())
        else:
            seri = serialize(dict(apiargs=args[i:], **kwds))
            noseriargs.append(seri if not_quoted else encode_str(seri))
            break
    else:
        # loop ended w/ break
        if kwds:
            seri = serialize(kwds)
            noseriargs.append(seri if not_quoted else encode_str(seri))
    # print(noseriargs)
    despaced = Serialize_Args_Sep.join(noseriargs)

    return despaced


def decode_str(a0):
    """
    """
    return urllib.parse.unquote(a0.replace(SAS_Avatar, Serialize_Args_Sep))


def deserialize_args(all_args, not_quoted=False, first_string=True, serialize_out=False):
    """ parse the command path to get positional and keywords arguments.

    1. if `not_quoted` is `True`, split everythine to the left of first `{` with `Serialize_Args_Sep` append the part startin from the `{`. `mark='{'`
    2. else after splitting all_args  with `Serialize_Args_Sep`: `mark='%7B%22'` (`quote('{')`)

    Scan from left. if all_args[i] not start with `mark`

    Conversion rules:
    |all_args[i]| converted to |
    | else | convert (case insensitive) and move on to the next segment |
    | string not starting with ```'0x'``` | `quote` |

    * else `decode_str()` if ```not_quoted==False``` else only substitute SAS_Avatar with Serialize_Args_Sep. Then `deserialize()` this segment to become ```{'apiargs':list, 'foo':bar ...}```, append value of ```apiargs``` to the converted-list above, remove the ```apiargs```-```val``` pair.
    * return 200 as the reurn code followed by the converted-list and the deserialized ```dict```.

    :all_args: a list of path segments for the args list.
    :first_string: Do not try to change the type of the first arg (assumed to be the function/method name).
    """
    args, kwds = [], {}

    if not_quoted:
        mark = '{'
        ar = all_args.split(mark, 1)
        qulist = ar[0].split(Serialize_Args_Sep)
        if len(ar) > 1:
            if len(qulist):
                # the last ',' was for mark so should be removed,
                qulist = qulist[:-1]
            qulist.append(mark + ar[1])
    else:
        mark = '%7B%22'
        qulist = all_args.split(Serialize_Args_Sep)
    # print(qulist)

    for a0 in qulist:
        if not a0.startswith(mark):
            # guess and change type
            # a string, bytes or number or boolean
            # if int(a0l.lstrip('+-').split('0x',1)[-1].isnumeric():
            # this covers '-/+0x34'
            if first_string:
                # do not try to change type
                arg = decode_str(a0)
                first_string = False  # do not come back here
            else:
                arg = guess_value(a0, last=decode_str)
            args.append(arg)
            # print(args)
        else:
            # quoted serialized dict
            readable = a0.replace(
                SAS_Avatar, Serialize_Args_Sep) if not_quoted else decode_str(a0)
            dese = deserialize(readable)
            if 'apiargs' in dese:
                args += dese['apiargs']
                del dese['apiargs']
            kwds = dese
            break
    logger.debug('args %s KWDS %s' % (str(args), str(kwds)))
    return 200, args, kwds



# -*- coding: utf-8 -*-

from .. import checkpath, PM_S
from ..model.user import auth

from ...dataset.deserialize import deserialize
from ...dataset.serializable import serialize
from ...dataset.mediawrapper import MediaWrapper
from ...pal.urn import makeUrn
from ...pal.poolmanager import PoolManager
from ...dataset.classes import Classes
from ...utils.common import (trbk,
                             getUidGid,
                             lls,
                             logging_ERROR,
                             logging_WARNING,
                             logging_INFO,
                             logging_DEBUG
                             )

from ...utils.fetch import fetch
from ...pns.fdi_requests import POST_PRODUCT_TAG_NAME

# from .db_utils import check_and_create_fdi_record_table, save_action

# import mysql.connector
# from mysql.connector import Error


from flask import request, make_response, Blueprint, current_app
from flask.wrappers import Response


import sys
import os
import time
import builtins
import importlib
import logging

if sys.version_info[0] >= 3:  # + 0.1 * sys.version_info[1] >= 3.3:
    PY3 = True
    strset = str
    from urllib.parse import urlparse
else:
    PY3 = False
    # strset = (str, unicode)
    strset = str
    from urlparse import urlparse


# Global variables set to temprary values before setGlabals() runs
logger = __import__('logging').getLogger(__name__)

data_api = Blueprint('httppool_server', __name__)

WRITE_LIST = ['POST', 'PUT', 'DELETE', 'PATCH']


# =============HTTP POOL=========================


def resp(code, result, msg, ts, serialize_out=False, ctype='application/json', length=80, req_auth=False):
    """
    Make response.

    :result: if is `Response`, this is returned req_auth applied.
    :ctype: Content-Type. Default is `application/json`
    :serialize_out: if True `result` is in already in serialized form.
    """

    # return if `result` is already a Response
    if issubclass(result.__class__, Response):
        resp = result
    else:
        if ctype == 'application/json':
            if serialize_out:
                # result is already in serialized form
                p = 'no-serialization-result-place-holder'
                t = serialize({"result": p, "msg": msg, "time": ts})
                w = t.replace('"'+p+'"', result)
            else:
                w = serialize({"result": result, "msg": msg, "time": ts})
        else:
            w = result

        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(lls(w, length))
            # logger.debug(pprint.pformat(w, depth=3, indent=4))

        resp = make_response(w, code)
        resp.headers['Content-Type'] = ctype

    if req_auth:
        resp.headers['WWW-Authenticate'] = 'Basic'
    return resp


def excp(e, code=400, msg='', serialize_out=True):
    # if current_app.config['DEBUG']:
    #    raise e
    result = '"FAILED"' if serialize_out else 'FAILED'
    msg = '%s\n%s: %s.\nTrace back: %s' % (
        msg, e.__class__.__name__, str(e), trbk(e))

    return code, result, msg


def check_readonly(usr, meth, logger):
    """ If both 'read_write' and 'read_only' present, 'read_write' has precedence.
    """
    return None
    if usr is None:
        msg = 'Unknown user %s.' % usr
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(msg)
        return msg, 401

    if meth in WRITE_LIST and ('read_only' in usr.roles) and ('read_write' not in usr.roles):
        msg = 'User %s is Read-Only, not allowed to %s.' % (usr.name, meth)
        if logger.isEnabledFor(logging_DEBUG):
            logger.debug(msg)
        return msg, 401

    return None


######################################
####  /urn{parts} get data ####
######################################


@ data_api.route('/urn<path:parts>', methods=['GET'])
@ auth.login_required
def urn(parts):
    """ Return data item from the given URN.

    :parts: parts of a URN, consists of the pool ID, a data class type, and a serial number (a.k.a index number). e.g. ``urn:pool:fdi.dataset.baseproduct.BaseProduct:0``, ``/pool/fdi.dataset.baseproduct.BaseProduct/0``. Also possible URL: ``http.../urn:pool/fdi.dataset.product.Product/0``".
    """

    serial_through = True
    logger = current_app.logger

    ts = time.time()
    logger.debug('get data for URN parts ' + parts)

    paths = parts2paths(parts)
    # if paths[-1] == '':
    #    del paths[-1]

    code, result, msg = getProduct_Or_Component(
        paths, serialize_out=serial_through)
    return resp(code, result, msg, ts, serialize_out=serial_through)

######################################
####  /urn{parts} remove data ####
######################################


@ data_api.route('/urn<path:parts>/', methods=['DELETE'])
@ data_api.route('/urn<path:parts>', methods=['DELETE'])
@ auth.login_required(role='read_write')
def delete_urn(parts):
    """ Remove data item with the given URN (or URN parts).

    :parts: parts of a URN, consists of the pool ID, a data class type, and a serial number (a.k.a index number). e.g. ``urn:pool:fdi.dataset.baseproduct.BaseProduct:0``, ``/pool/fdi.dataset.baseproduct.BaseProduct/0``. Also possible URL: ``http.../urn:pool/fdi.dataset.product.Product/0``".
    """

    serial_through = True
    logger = current_app.logger

    ts = time.time()
    logger.info('delete data for URN parts ' + parts)

    paths = parts2paths(parts)
    # if paths[-1] == '':
    #    del paths[-1]

    code, result, msg = delete_product(paths, serialize_out=False)
    return resp(code, result, msg, ts, serialize_out=False)


def parts2paths(parts):
    # parts == urn:{type}:{index}/...
    # parts == /urn:{type}:{index}/...
    # parts == :{type}:{index}/...
    # parts == /{type}:{index}/...
    # parts == /{type}/{index}/...
    if parts[:4].lower() == 'urn:':
        parts = parts[4:]
    elif parts[:5].lower() == '/urn:':
        parts = parts[5:]
    elif parts[0] == '/' or parts[0] == ':':
        parts = parts[1:]
    # deal with the possible ':' before index
    sp1 = parts.split('/')
    if ':' in sp1[0]:
        paths = sp1[0].split(':') + sp1[1:]
    else:
        paths = sp1
    return paths


def delete_product(paths, serialize_out=False):
    """ removes specified product from pool
    """
    FAILED = '"FAILED"' if serialize_out else 'FAILED'

    typename = paths[1]
    indexstr = paths[2]
    poolname = paths[0]
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    urn = makeUrn(poolname=poolname, typename=typename, index=indexstr)
    # resourcetype = fullname(data)

    if not PM_S.isLoaded(poolname):
        result = FAILED
        msg = 'Pool not found or not registered: ' + poolname
        code = 404
        logger.error(msg)
        return code, result, msg
    logger.debug('DELETE product urn: ' + urn)
    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        logger.debug('**** '+str(list(poolobj._dTypes))+urn)
        poolobj.remove(urn)
        result = 0
        msg = 'remove product ' + urn + ' OK.'
        code = 200
    except Exception as e:
        code, result, msg = excp(
            e,
            msg='Unable to remove product: ' + urn)
        logger.error(msg)
    return code, result, msg

######################################
####  {pool}/ POST   ####
######################################


@ data_api.route('/<string:pool>/', methods=['POST'])
@ data_api.route('/<string:pool>', methods=['POST'])
@ auth.login_required(role='read_write')
def save_data(pool):
    """
    Save data to the pool with a list of tags and receive URNs.

    Save product data item(s) to the pool with an optional set of tags (The same tags are given to every data item) and receive a URN for each of the saved items.
    """

    ts = time.time()
    logger = current_app.logger
    # logger.debug(f'save to ' + pool)

    # do not deserialize if set True. save directly to disk
    serial_through = True

    if not request.data:
        result, msg = '"FAILED"', 'No Request data for command '+request.method
        code = 400
        logger.warning(f'SAVE product to {pool} {msg}.')
        return resp(code, result, msg, ts, serialize_out=True)

    # save product
    if request.headers.get(POST_PRODUCT_TAG_NAME) is not None:
        tags = deserialize(request.headers.get(POST_PRODUCT_TAG_NAME))
    else:
        tags = None

    paths = [pool]
    # logger.debug('*** method:%s pool:%s tags:%s' %
    #             (request.method, pool, str(tags)))

    if serial_through:
        data = str(request.data, encoding='ascii')

        code, result, msg = save_product(
            data, paths, tags, serialize_in=not serial_through, serialize_out=serial_through, logger=logger)
    else:
        try:
            data = deserialize(request.data)
        except ValueError as e:
            code, result, msg = excp(
                e,
                msg='Class needs to be included in pool configuration.',
                serialize_out=serial_through)
            logger.info(f'SAVE product to {poolurl} {msg}.')
        else:
            code, result, msg = save_product(
                data, paths, tags, serialize_in=not serial_through, logger=logger)
            # save_action(username=username, action='SAVE', pool=paths[0])

    return resp(code, result, msg, ts, serialize_out=serial_through)


def save_product(data, paths, tags=None, serialize_in=True, serialize_out=False, logger=logger):
    """Save products and returns URNs.

    Saving Products to HTTPpool will have data stored on the server side. The server only returns URN strings as a response. ProductRefs will be generated by the associated httpclient pool which is the front-end on the user side.

    :tags: a list of tag strings. default is None meaning no tag.
    Returns a URN object or a list of URN objects.
    """
    FAILED = '"FAILED"' if serialize_out else 'FAILED'

    poolname = paths[0]
    fullpoolpath = os.path.join(
        current_app.config['FULL_BASE_LOCAL_POOLPATH'], poolname)
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    # resourcetype = fullname(data)

    if checkpath(fullpoolpath, current_app.config['PC']['self_username']) is None:
        result = FAILED
        msg = 'Pool directory error: ' + fullpoolpath
        logger.info(f'SAVE product to {poolurl} {msg}.')
        return 400, result, msg

    # logger.debug(str(id(PM_S._GlobalPoolList)) + ' ' + str(PM_S._GlobalPoolList))

    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.saveProduct(
            product=data, tag=tags, geturnobjs=True, serialize_in=serialize_in, serialize_out=serialize_out)
        # seriaized:'"urn:test_remote_pool:fdi.dataset.product.Product:0"'
        urn = result.split(poolname, 1)[1][1:-1] if serialize_out else result
        msg = f'Save {urn} to {poolurl} OK.'
        code = 200
        logger.info(msg)
    except Exception as e:
        code, result, msg = excp(e, serialize_out=serialize_out)
        logger.info(f'SAVE product to {poolurl} {msg}.')
    return code, result, msg

######################################
####  {pool}/{data_paths}  GET  ####
######################################


@ data_api.route('/<string:pool>/<path:data_paths>', methods=['GET'])
@ auth.login_required(role='read_write')
def data_paths(pool, data_paths):
    """
    Returns magics of given type/data in the given pool.


    """

    ts = time.time()
    logger = current_app.logger
    serial_through = True

    # do not deserialize if set True. save directly to disk

    paths = [pool] + parts2paths(data_paths)

    logger.debug('>>>[%4s] %s data_paths= %s paths= %s' %
                 (request.method, pool, str(data_paths), str(paths)))

    code, result, msg = getProduct_Or_Component(
        paths, serialize_out=serial_through)
    return resp(code, result, msg, ts, serialize_out=serial_through)


def getProduct_Or_Component(paths, serialize_out=False):
    """
    :serialize_out: see :meth:`ProductPool.saveProduct`
    """

    lp = len(paths)
    # now paths = poolname, prod_type , ...
    logger.debug('get prod or compo: ' + str(paths))

    ts = time.time()
    mInfo = 0
    if lp == 2:
        # ex: pool/fdi.dataset.Product
        # return classes[class]
        pp = paths[1]
        mp = pp.rsplit('.', 1)
        if len(mp) < 2:
            msg = 'Need a dot-separated full type name, not %s.' % pp
            code = 422
            logger.info(f'{code} {msg}')
            return code, '"FAILED"', msg
        modname, ptype = mp[0], mp[1]
        cls = Classes.mapping[ptype]
        mod = importlib.import_module(modname)  # TODO
        try:
            mInfo = getattr(mod, 'Model')
        except AttributeError:
            mInfo = cls().zInfo
        # non-serialized
        code = 200
        msg = 'Getting API info for %s OK' % paths[1]
        logger.info(f'{code} {msg}')
        return 0, resp(code, mInfo,  msg, ts, serialize_out=False), 0
    # elif lp == 3 and paths[-1]=='':

    #     try:
    #         poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
    #         result = poolobj.readHK(hkname, serialize_out=serialize_out)
    #         code, msg = 200, hkname + ' HK data returned OK'
    #     except Exception as e:
    #         code, result, msg = excp(e, serialize_out=serialize_out)
    elif lp >= 3:
        return get_component_or_method(paths, mInfo, serialize_out=serialize_out)

    else:
        __import__("pdb").set_trace()

        code = 400
        msg = 'Unknown path %s' % str(paths)
        logger.info(f'{code} {msg}')
        return code, '"FAILED"', msg


HTML_STYLE = """
<style>
table,th,td {
    border: 1px solid black;
    border-collapse: collapse;
}
tr:nth-child(even){
    background-color: #DFDFDF;
    color: black;
}
</style>
"""


def get_component_or_method(paths, mInfo, serialize_out=False):
    """ Get the component and the associated command and return

    Except for full products, most components  are not in serialized form.
    """
    FAILED = '"FAILED"' if serialize_out else 'FAILED'
    ts = time.time()
    logger.debug('get compo or meth: ' + str(paths))
    lp = len(paths)
    # if paths[-1] in ('toString', 'string'):
    #    __import__('pdb').set_trace()

    if paths[-1] == '':
        # command is '' and url endswith a'/'
        compo, path_str, prod = load_component_at(1, paths[:], mInfo)
        if compo is not None:
            ls = [m for m in dir(compo) if not m.startswith('_')]
            return 0, resp(200, ls, 'Getting %s members/attrbutes OK' % (path_str),
                           ts, serialize_out=False), 0
        else:
            return 400, FAILED, 'Cannot get %s for "%s".' % ('/'.join(paths[:-1]), path_str)
    elif lp == 3:
        # url ends with index
        # no cmd, ex: test/fdi.dataset.Product/4
        # send json of the prod

        code, result, msg = load_product(1, paths, serialize_out=serialize_out)
        return 0, resp(code, result, msg, ts, serialize_out=serialize_out), 0
    elif paths[2].isnumeric():
        # grand tour
        compo, path_str, prod = load_component_at(1, paths, mInfo)
        # see :func:`fetch`
        # e.g. path_str is like '.string()' '["text"]'  '.meta["speed"].isValid()'
        if compo or 'has no' not in path_str:
            code = 200
            msg = f'Getting {path_str} OK'
            compo_meth_name = path_str.split('.')[-1]
            if compo_meth_name[:8] == 'toString' or \
               compo_meth_name[:6] == 'string' or \
               compo_meth_name[:3] == 'txt':
                if 'html' in compo_meth_name:
                    ct = 'text/html'
                    result = HTML_STYLE + compo
                elif 'rst' in compo_meth_name:
                    ct = 'text/plain;charset=utf-8'
                    result = compo
                elif 'fancy_grid' in compo_meth_name:
                    ct = 'text/plain;charset=utf-8'
                    result = compo
                else:
                    ct = 'text/plain'
                    result = compo
                return 0, resp(code, result, msg, ts, ctype=ct, serialize_out=False), 0
            elif compo_meth_name[:4] == 'html':
                ct = 'text/html'
                result = HTML_STYLE + compo
                return 0, resp(code, result, msg, ts, ctype=ct, serialize_out=False), 0
            elif compo_meth_name.startswith('yaml(') or compo_meth_name.startswith('tree('):
                ct = 'text/plain;charset=utf-8'
                # 'font-family: "Courier New",monospace;\n'
                result = compo
                return 0, resp(code, result, msg, ts, ctype=ct, serialize_out=False), 0
            elif compo_meth_name.startswith('fits('):
                ct = 'application/fits'
                result = compo
                return 0, resp(code, result, msg, ts, ctype=ct, serialize_out=False), 0
            elif issubclass(compo.__class__, MediaWrapper):
                ct = compo.type
                result = compo.data

                return 0, resp(code, result, msg, ts, ctype=ct, serialize_out=False), 0
            elif compo_meth_name.startswith('graph(') or \
                    compo_meth_name.startswith('getTaskHistory('):
                ct = 'image/png'
                result = compo
                return 0, resp(code, result, msg, ts, ctype=ct, serialize_out=False), 0

            else:
                return 0, resp(code, compo, msg, ts, serialize_out=False), 0

        else:
            return 400, FAILED, '%s: %s' % (str(paths), path_str)

    elif 0:
        # no cmd, ex: test/fdi.dataset.Product/4
        # send json of the prod component
        compo, path_str, prod = load_component_at(1, paths, mInfo)
        # see :func:`fetch`
        if compo or ' non ' not in path_str:
            return 0, resp(
                200, compo,
                'Getting %s OK' % (cmd + ':' + paths[2] + '/' + path_str),
                ts, serialize_out=False), 0
        else:
            return 400, FAILED, '%s : %s' % ('/'.join(paths[:3]), path_str)
    else:
        return 400, FAILED, 'Need index number %s' % str(paths)


def load_component_at(pos, paths, mInfo):
    """ paths[pos] is data_type; paths[pos+2] is 'description', 'meta' ...

    Components fetched are not in serialized form.
    """
    # component = fetch(paths[pos+2:], mInfo)
    # if component:

    # get the product live
    code, live_prod, msg = load_product(pos, paths, serialize_out=False)
    if code != 200:
        return None, '%s. Unable to load %s.' % (msg, str(paths)), None
    compo, path_str = fetch(paths[pos+2:], live_prod,
                            exe=['*', 'is', 'get'], not_quoted=False)

    return compo, path_str, live_prod


def load_product(p, paths, serialize_out=False):
    """Load product paths[p]: paths[p+1] from paths[0]
    """
    FAILED = '"FAILED"' if serialize_out else 'FAILED'

    typename = paths[p]
    indexstr = paths[p+1]
    poolname = paths[0]
    poolurl = current_app.config['POOLURL_BASE'] + poolname
    urn = makeUrn(poolname=poolname, typename=typename, index=indexstr)
    # resourcetype = fullname(data)

    logger.debug('LOAD product: ' + urn)
    try:
        poolobj = PM_S.getPool(poolname=poolname, poolurl=poolurl)
        result = poolobj.loadProduct(urn=urn, serialize_out=serialize_out)
        msg = ''
        code = 200
    except Exception as e:
        if issubclass(e.__class__, NameError):
            msg = 'Not found: ' + poolname
            code = 404
        else:
            __import__("pdb").set_trace()

            msg, code = '', 400
        code, result, msg = excp(
            e, code=code, msg=msg, serialize_out=serialize_out)
    return code, result, msg


Builtins = vars(builtins)


def mkv(v, t):
    """
    return v with a tyoe specified by t.

    t: 'NoneType' or any name in ``Builtins``.
    """

    m = v if t == 'str' else None if t == 'NoneType' else Builtins[t](
        v) if t in Builtins else deserialize(v)
    return m



# -*- coding: utf-8 -*-

#from fdi.pal.poolmanager import PoolManager
#from fdi.pns.jsonio import getJsonObj, postJsonObj, putJsonObj, commonheaders
#from fdi.utils import getconfig
#from fdi.utils.common import lls
#
#import pytest
#import os
#import logging
##logger = logging.getLogger(__name__)
#




# -*- coding: utf-8 -*-
import logging
from weakref import WeakValueDictionary, getweakrefcount
import getpass

from ..utils.getconfig import getConfig
from ..utils.common import lls
from .urn import parse_poolurl
from ..pal.httppool import HttpPool

from requests.exceptions import ConnectionError
import requests

# create logger
logger = logging.getLogger(__name__)
# logger.debug('level %d' %  (logger.getEffectiveLevel()))

pc = getConfig()

DEFAULT_MEM_POOL = 'defaultmem'
# localpool
DEFAULT_POOL = 'fdi_pool_' + __name__ + getpass.getuser()
Invalid_Pool_Names = ['pools', 'urn', 'URN', 'api']


def remoteRegister(pool):
    """ if registered a pool's auth and client will be used.

    Note that a new http/csdb pool gets remoteRegistered before locally registered.

    Parameter
    ---------
    pool : HttpClientPool, PublicClientPool
        Pool object to be registered on remote server and have client/session set.
    auth : object
        Authorization object for the client. If given will substitute that of pool, if pool has auth.
    client : flask.requests (testing), or requests.Session
        The client. If given will substitute that of pool, if pool is given

    """
    # pool object
    poolo = None
    from ..pal import httpclientpool, publicclientpool

    if issubclass(pool.__class__, httpclientpool.HttpClientPool):

        # HttpClientPool. Init the remote pool. If exists, load.d
        poolurl = pool._poolurl
        poolo = pool

        logger.debug('Register %s on the server', poolurl)
        if poolurl.endswith('/'):
            poolurl = poolurl[:-1]
        from ..pns.fdi_requests import put_on_server
        try:
            res, msg = put_on_server(
                'urn:::0', poolurl, 'register_pool', auth=poolo.auth, client=poolo.client)
        except ConnectionError as e:
            res, msg = 'FAILED', str(e)
            logger.error(poolurl + ' ' + msg)
            raise
        if res == 'FAILED':
            np = '<' + poolo.auth.username + ' ' + poolo.auth.password + \
                '>' if poolo.auth else '<no authorization>'
            raise RuntimeError(
                'Registering ' + poolurl + ' failed with auth ' + np + ' , ' + msg)
        return res, msg
    elif issubclass(pool.__class__, publicclientpool.PublicClientPool):
        from ..pns.fdi_requests import ServerError
        # register csdb pool. If existing, load. IF not exists, create and initialize sn.
        poolurl = pool._poolurl
        try:
            res = pool.createPool2()

            pool.getToken()
            pool.client.headers.update({'X-AUTH-TOKEN': pool.token})
            msg = 'New pool made.'
            pool.poolInfo = pool.getPoolInfo()
            res = 'OK'
        except ServerError as e:
            # if e.code == 1:
            #   msg = 'Pool exists or bad namw.'
            msg = str(e)
            logger.error(
                'Registering ' + poolurl + ' failed with auth ' + 'np' + ' , ' + msg)
            res = 'FAILED'
            raise
        return res, msg
    else:
        return


def remoteUnregister(poolurl, auth=None, client=None):
    """ this method does not reference pool object. """

    if not poolurl.lower().startswith('http'):
        logger.warning('Ignored: %s not for a remote pool.' % poolurl)
        return 1
    logger.debug('unregister %s on the server', poolurl)

    # check if poolurl has been registered
    for pool, poolo in PoolManager._GlobalPoolList.items():
        if issubclass(poolo.__class__, HttpPool):
            continue
        if poolurl == poolo._poolurl:
            if client is None:
                client = poolo.client
            if auth is None:
                auth = poolo.auth
            break
    else:
        raise NameError(
            f'Remote Unregistering failed. {poolurl} not registered or not suitable.')
    from ..pns.fdi_requests import delete_from_server
    # url = api_baseurl + post_poolid
    # x = requests.delete(url, auth=HTTPBasicAuth(auth_user, auth_pass))
    # o = deserialize(x.text)
    urn = 'urn:::0'
    try:
        res, msg = delete_from_server(
            urn, poolurl, 'unregister_pool', auth=auth, client=client)
    except ConnectionError as e:
        res, msg = 'FAILED', str(e)
    if res == 'FAILED':
        msg = f'Unregistering {poolurl} failed. {msg}'
        if getattr(poolo, 'ignore_error_when_delete', False):
            logger.info('Ignored: ' + msg)
            code = 2
        else:
            raise ValueError(msg)
    else:
        code = 0
    return code


class PoolManager(object):
    """
    This class provides the means to reference ProductPool objects without having to hard-code the type of pool. For example, it could be desired to easily switch from one pool type to another.

This is done by calling the getPool() method, which will return an existing pool or create a new one if necessary.
    """

    _GlobalPoolList = WeakValueDictionary()
    """ Global centralized dict that returns singleton -- the same -- pool for the same ID."""

    # maps scheme to default place/poolpath
    # pc['host']+':'+str(pc['port'])+pc['baseurl']
    p = getConfig('poolurl:').strip('/').split('://')[1]
    PlacePaths = {
        'file': pc['base_local_poolpath'],
        'mem': '/',
        'http': p,
        'https': p,
        'server': pc['server_local_poolpath'],
        'csdb': pc['cloud_api_version']
    }
    del p

    @classmethod
    def getPool(cls, poolname=None, poolurl=None, pool=None, makenew=True, auth=None, client=None, **kwds):
        """ returns an instance of pool according to name or path of the pool.

        Returns the pool object if the pool is registered. Creates the pool if it does not already exist. the same poolname-path always get the same pool. Http pools will be registered on the server side.

Pools registered are kept as long as the last reference remains. When the last is gone the pool gets :meth;`removed` d.

        Parameter
        ---------
        poolname : str
            name of the pool.
        poolurl : str
            If given the poolpath, scheme, place will be derived from it. if not given for making a new pool (i.e. when poolname is not a registered pool name. If poolname is missing it is derived from poolurl; if poolurl is also absent, ValueError will be raised.
        pool: ProductPool
            If `auth` and `client` are given they will substitute those of  `pool`. If `pool` is not given, those will need to be given.
        makenew : bool
            When the pool does not exist, make a new one (````True```; default) or throws `PoolNotFoundError` (```False```).
        auth : str
            For `remoteRegister`.
        client : default is `None`.
            For `remoteRegister`.
        kwds  : dict
            Passed to pool instanciation arg-list.

        Returns
        -------
        ProductPool:
            The pool object.
        """
        # logger.debug('GPL ' + str(id(cls._GlobalPoolList)) +
        #             str(cls._GlobalPoolList) + ' PConf ' + str(cls.PlacePaths))
        if pool:
            if poolname:
                raise ValueError(
                    'Pool name %s and pool object cannot be both given.' % poolname)
            poolname, poolurl = pool._poolname, pool._poolurl
            if cls.isLoaded(poolname):
                return cls._GlobalPoolList[poolname]
            if poolurl.lower()[:4] in ('http', 'csdb'):
                if auth is not None and getattr(pool, 'auth', None) is None:
                    pool.auth = auth
                if client is not None and getattr(pool, 'client', None) is None:
                    from ..httppool.session import requests_retry_session
                    pool.client = requests_retry_session()
                res, msg = remoteRegister(pool)
            p = pool
        else:
            # quick decisions can be made knowing poolname only
            if poolname == DEFAULT_MEM_POOL:
                if not poolurl:
                    poolurl = 'mem:///' + poolname
            if poolname is not None:
                if poolname in Invalid_Pool_Names:
                    raise ValueError(
                        'Cannot register invalid pool name: ' + poolname)
                if cls.isLoaded(poolname):
                    return cls._GlobalPoolList[poolname]

            # get poolname and scheme
            if not poolurl:
                poolurl = getConfig('poolurl:'+poolname)
            if poolurl:
                pp, schm, pl, pn, un, pw = parse_poolurl(poolurl)
            else:
                raise ValueError(
                    'A new pool %s cannot be created without a pool url. Maybe the pool needs to be registered?' % poolname)
            if poolname:
                if pn != poolname:
                    raise ValueError(
                        f'Poolname in poolurl {poolurl} is different from poolname {poolname}.')
            else:
                poolname = pn

            # now we have scheme, poolname, poolurl
            if poolname in Invalid_Pool_Names:
                raise ValueError(
                    'Cannot register invalid pool name: ' + poolname)
            if cls.isLoaded(poolname):
                return cls._GlobalPoolList[poolname]
            if schm == 'file':
                from . import localpool
                p = localpool.LocalPool(
                    poolname=poolname, poolurl=poolurl, makenew=makenew, **kwds)
            elif schm == 'mem':
                from . import mempool
                p = mempool.MemPool(poolname=poolname, poolurl=poolurl, **kwds)
            elif schm == 'server':
                from . import httppool
                p = httppool.HttpPool(
                    poolname=poolname, poolurl=poolurl, **kwds)
            elif schm in ('http', 'https', 'csdb'):
                if schm == 'csdb':
                    from . import publicclientpool
                    pooltype = publicclientpool.PublicClientPool
                    purl = pc['scheme'] + poolurl[4:]
                else:
                    from . import httpclientpool
                    pooltype = httpclientpool.HttpClientPool
                    purl = poolurl
                if auth is None:
                    from requests.auth import HTTPBasicAuth
                    auth = HTTPBasicAuth(pc['username'],
                                         pc['password'])
                if client is None:
                    from ..httppool.session import requests_retry_session
                    client = requests_retry_session()
                p = pooltype(poolname=poolname, poolurl=purl,
                             auth=auth, client=client, **kwds)
                res, msg = remoteRegister(p)
            else:
                raise NotImplementedError(schm + ':// is not supported')
        # print(getweakrefs(p), id(p), '////')
        cls.save(poolname, p)
        # print(getweakrefs(p), id(p))

        # Pass poolurl to PoolManager.remove() for remote pools
        # finalize(p, print, poolname, poolurl)
        logger.debug('made pool ' + lls(p, 900))
        return p

    @ classmethod
    def getMap(cls):
        """
        Returns a poolname - poolobject map.
        """
        return cls._GlobalPoolList

    @ classmethod
    def isLoaded(cls, poolname):
        """
        Whether an item with the given id has been loaded (cached).

        :returns: the number of remaining week references if the pool is loaded. Returns 0 if poolname is not found in _GlobalPoolList or weakref count is 0.
        """
        if poolname in cls._GlobalPoolList:
            # print(poolname, getweakrefcount(cls._GlobalPoolList[poolname]))
            return getweakrefcount(cls._GlobalPoolList[poolname])
        else:
            return 0

    @ classmethod
    def removeAll(cls, ignore_error=False):
        """ deletes all pools from the pool list, pools not wiped
        """
        nl = list(cls._GlobalPoolList)
        for pool in nl:
            cls.remove(pool, ignore_error=ignore_error)

    @ classmethod
    def save(cls, poolname, poolobj):
        """
        """
        cls._GlobalPoolList[poolname] = poolobj
        poolobj.setPoolManager(cls)

    @ classmethod
    def remove(cls, poolname, ignore_error=False):
        """ Remove from list and unregister remote pools.

        returns 0 for successful removal, ``1`` for poolname not registered or referenced, still attempted to remove. ``> 1`` for the number of weakrefs the pool still have, and removing failed.
        """
        # number of weakrefs
        nwr = cls.isLoaded(poolname)
        # print(getweakrefs(cls._GlobalPoolList[poolname]), id(
        #    cls._GlobalPoolList[poolname]), '......', nwr)

        if nwr == 1:
            # this is the only reference. unregister remote first.
            thepool = cls._GlobalPoolList[poolname]
            poolurl = thepool._poolurl
            from ..pal.httpclientpool import HttpClientPool
            if issubclass(thepool.__class__, HttpClientPool):
                code = remoteUnregister(poolurl)
            else:
                code = 0
        elif nwr > 1:
            # nothing needs to be done. weakref number will decrement after Storage deletes ref
            return nwr
        else:
            # nwr <=  0
            code = 1
        try:
            pool = cls._GlobalPoolList.pop(poolname)
            pool.setPoolManager(None)
        except KeyError as e:
            if ignore_error:
                logger.info("Ignored: "+str(e))
            else:
                raise
        return code

    @ classmethod
    def getPoolurlMap(cls):
        """
        Gives the default poolurls of PoolManager.
        """
        return cls.PlacePaths

    @ classmethod
    def setPoolurlMap(cls, new):
        """
        Sets the default poolurls of PoolManager.
        """
        cls.PlacePaths.clear()
        cls.PlacePaths.update(new)

    @ classmethod
    def size(cls):
        """
        Gives the number of entries in this manager.
        """
        return len(cls._GlobalPoolList)

    def items(self):
        """
        Returns map's items
        """
        return self._GlobalPoolList.items()

    def __setitem__(self, poolname, poolobj):
        """ sets value at key.
        """
        self._GlobalPoolList.__setitem__(poolname, poolobj)
        poolobj.setPoolManager(None, self.__class__)

    def __getitem__(self, *args, **kwargs):
        """ returns value at key.
        """
        return self._GlobalPoolList.__getitem__(*args, **kwargs)

    def __delitem__(self, poolname):
        """ removes value and its key.
        """
        self._GlobalPoolList[poolname].setPoolManager(None)
        self._GlobalPoolList.__delitem__(poolname)

    def __len__(self, *args, **kwargs):
        """ size of data
        """
        return self._GlobalPoolList.__len__(*args, **kwargs)

    def __iter__(self, *args, **kwargs):
        """ returns an iterator
        """
        return self._GlobalPoolList.__iter__(*args, **kwargs)

    def __repr__(self):
        return self.__class__.__name__ + '(' + str(self._GlobalPoolList) + ')'
